#ifndef INTERNAL_LOADPATH_H                              /*-*-C-*-vi:se ft=c:*/
#define INTERNAL_LOADPATH_H
/**
 * @author     Ruby developers <ruby-core@ruby-lang.org>
 * @copyright  This  file  is   a  part  of  the   programming  language  Ruby.
 *             Permission  is hereby  granted,  to  either redistribute  and/or
 *             modify this file, provided that  the conditions mentioned in the
 *             file COPYING are met.  Consult the file for details.
 * @brief      Internal header for $LOAD_PATH.
 */

/* loadpath.c */
extern const char ruby_exec_prefix[];
extern const char ruby_initial_load_paths[];

#endif /* INTERNAL_LOADPATH_H */
